/*
 *  stereo_match.cpp
 *  calibration
 *
 *  Created by Victor  Eruhimov on 1/18/10.
 *  Copyright 2010 Argus Corp. All rights reserved.
 * -----------------------------------------------------------------
 *  Modified to capture input from two cameras
 *  Giandomenico De Sanctis @ 2/2012
 */

#include "opencv2/calib3d/calib3d.hpp"
#include "opencv2/imgproc/imgproc.hpp"
#include "opencv2/imgproc/imgproc_c.h"
#include "opencv2/highgui/highgui.hpp"
#include "opencv2/contrib/contrib.hpp"

#include "opencv2/legacy/legacy.hpp"


#include <stdio.h>
#include <iostream>

using namespace cv;

void print_help()
{
	printf("\nDemo stereo matching converting L and R images into disparity and depth map\n");
    printf("\nUsage: stereo_match  <left_camera_id>  <right_camera_id> [--algorithm=bm|sgbm|hh|var|bmleg] [--blocksize=<block_size>]\n"
           "[--max-disparity=<max_disparity>] [--scale=<scale_factor>] [-i <intrinsic_filename>] [-e <extrinsic_filename>]\n"
           "[-p]\n");
	printf("-p  display depth map\n");
}

void saveXYZ(const char* filename, const Mat& mat)
{
    const double max_z = 1.0e4;
    FILE* fp = fopen(filename, "wt");
    for(int y = 0; y < mat.rows; y++)
    {
        for(int x = 0; x < mat.cols; x++)
        {
            Vec3f point = mat.at<Vec3f>(y, x);
            if(fabs(point[2] - max_z) < FLT_EPSILON || fabs(point[2]) > max_z) continue;
            fprintf(fp, "%f %f %f\n", point[0], point[1], point[2]);
        }
    }
    fclose(fp);
}

int main(int argc, char** argv)
{
    const char* algorithm_opt = "--algorithm=";
    const char* maxdisp_opt = "--max-disparity=";
    const char* blocksize_opt = "--blocksize=";
    const char* scale_opt = "--scale=";
    
    if(argc < 3)
    {
        print_help();
		return 0;
    }
    const char* img1_filename = 0;
    const char* img2_filename = 0;
    
    int  leftCam = -1;
    int  rightCam = -1;

    const char* intrinsic_filename = 0;
    const char* extrinsic_filename = 0;
    const char* disparity_filename = 0;
    const char* point_cloud_filename = 0;
    
    enum { STEREO_BM=0, STEREO_SGBM=1, STEREO_HH=2, STEREO_VAR=3, STEREO_LEGACY=4 };
    int alg = STEREO_SGBM;
    int SADWindowSize = 0, numberOfDisparities = 0;
    bool no_display = false;
    float scale = 1.f;
    
    StereoBM bm;
    StereoSGBM sgbm;
    StereoVar var;
    
    for( int i = 1; i < argc; i++ )
    {
        if( argv[i][0] != '-' )
        {
            if( leftCam == -1 )
                leftCam = atoi(argv[i]);
            else
                rightCam = atoi(argv[i]);
        }
        else if( strncmp(argv[i], algorithm_opt, strlen(algorithm_opt)) == 0 )
        {
            char* _alg = argv[i] + strlen(algorithm_opt);
            alg = strcmp(_alg, "bm") == 0 ? STEREO_BM :
                  strcmp(_alg, "sgbm") == 0 ? STEREO_SGBM :
                  strcmp(_alg, "hh") == 0 ? STEREO_HH :
                  strcmp(_alg, "var") == 0 ? STEREO_VAR : 
                  strcmp(_alg, "bmleg") == 0 ? STEREO_LEGACY : -1;
            if( alg < 0 )
            {
                printf("Command-line parameter error: Unknown stereo algorithm\n\n");
                print_help();
                return -1;
            }
        }
        else if( strncmp(argv[i], maxdisp_opt, strlen(maxdisp_opt)) == 0 )
        {
            if( sscanf( argv[i] + strlen(maxdisp_opt), "%d", &numberOfDisparities ) != 1 ||
                numberOfDisparities < 1 || numberOfDisparities % 16 != 0 )
            {
                printf("Command-line parameter error: The max disparity (--maxdisparity=<...>) must be a positive integer divisible by 16\n");
                print_help();
                return -1;
            }
        }
        else if( strncmp(argv[i], blocksize_opt, strlen(blocksize_opt)) == 0 )
        {
            if( sscanf( argv[i] + strlen(blocksize_opt), "%d", &SADWindowSize ) != 1 ||
                SADWindowSize < 1 || SADWindowSize % 2 != 1 )
            {
                printf("Command-line parameter error: The block size (--blocksize=<...>) must be a positive odd number\n");
                return -1;
            }
        }
        else if( strncmp(argv[i], scale_opt, strlen(scale_opt)) == 0 )
        {
            if( sscanf( argv[i] + strlen(scale_opt), "%f", &scale ) != 1 || scale < 0 )
            {
                printf("Command-line parameter error: The scale factor (--scale=<...>) must be a positive floating-point number\n");
                return -1;
            }
        }
        else if( strcmp(argv[i], "-i" ) == 0 )
            intrinsic_filename = argv[++i];
        else if( strcmp(argv[i], "-e" ) == 0 )
            extrinsic_filename = argv[++i];
        else if( strcmp(argv[i], "-o" ) == 0 )
            disparity_filename = argv[++i];
        else if( strcmp(argv[i], "-p" ) == 0 )
	   point_cloud_filename = "ok";
        else
        {
            printf("Command-line parameter error: unknown option %s\n", argv[i]);
            return -1;
        }
    }
    
    if ((leftCam == -1) || (rightCam == -1))	
    {
        printf("Command-line parameter error: both left and right cameras numeric ID must be specified\n");
        return -1;
    }
    
    if( (intrinsic_filename != 0) ^ (extrinsic_filename != 0) )
    {
        printf("Command-line parameter error: either both intrinsic and extrinsic parameters must be specified, or none of them (when the stereo pair is already rectified)\n");
        return -1;
    }
    
    if( extrinsic_filename == 0 && point_cloud_filename )
    {
        printf("Command-line parameter error: extrinsic and intrinsic parameters must be specified to compute the point cloud\n");
        return -1;
    }
    
    int color_mode = alg == STEREO_BM ? 0 : -1;

    Mat img1 ; 
    Mat img2 ; 
    Mat map11, map12, map21, map22;
    Mat img1r, img2r;
    Mat temp1, temp2;
    Size img_size;    
    
    Mat M1, D1, M2, D2;
    Mat R, T, R1, P1, R2, P2;
    Mat Q;
	const double MAX_DEPTH_Z = 10000.0;

	// left
    CvCapture* captureL = cvCaptureFromCAM( leftCam );
    assert( captureL );

    // right
    CvCapture* captureR = cvCaptureFromCAM( rightCam );
    assert( captureR );

	//cvSetCaptureProperty(captureL, CV_CAP_PROP_FRAME_WIDTH, 320.0);
	//cvSetCaptureProperty(captureL, CV_CAP_PROP_FRAME_HEIGHT, 240.0);

	//cvSetCaptureProperty(captureR, CV_CAP_PROP_FRAME_WIDTH, 320.0);
	//cvSetCaptureProperty(captureR, CV_CAP_PROP_FRAME_HEIGHT, 240.0);

    // store in buffer the frames from the 2 cameras using the fast function cvGrabFrame,
    // so the frames have the best synchronization
    cvGrabFrame( captureL );
    cvGrabFrame( captureR );

    // retrieve the images from buffer by the (relatively) slow function cvRetrieveFrame	
    IplImage *imageL = cvRetrieveFrame( captureL );
    IplImage *imageR = cvRetrieveFrame( captureR );

    IplImage *grayL = cvCreateImage(cvGetSize(imageL), 8, 1);
    IplImage *grayR = cvCreateImage(cvGetSize(imageL), 8, 1);

    IplImage *laplaceOut = cvCreateImage(cvGetSize(imageL), IPL_DEPTH_32F, 1);

	cvCvtColor(imageL, grayL, CV_BGR2GRAY);
	cvCvtColor(imageR, grayR, CV_BGR2GRAY);

    img1 = grayL;
    img2 = grayR;

	Mat tempMat = Mat(img1.cols, img1.rows, CV_8UC1);

   	img_size = img1.size();
    Rect roi1 = Rect(cvRect(0,0, img1.cols, img1.rows ));
	Rect roi2 = Rect(cvRect(0,0, img1.cols, img1.rows ));

    if( intrinsic_filename )
    {
        // reading intrinsic parameters
        FileStorage fs(intrinsic_filename, CV_STORAGE_READ);
        if(!fs.isOpened())
        {
            printf("Failed to open file %s\n", intrinsic_filename);
		    cvReleaseCapture(&captureL);
			cvReleaseCapture(&captureR);
            return -1;
        }
        
        fs["M1"] >> M1;
        fs["D1"] >> D1;
        fs["M2"] >> M2;
        fs["D2"] >> D2;
        
        fs.open(extrinsic_filename, CV_STORAGE_READ);
        if(!fs.isOpened())
        {
            printf("Failed to open file %s\n", extrinsic_filename);
		    cvReleaseCapture(&captureL);
			cvReleaseCapture(&captureR);
           return -1;
        }
        
        fs["R"] >> R;
        fs["T"] >> T;
        
        //fs["R1"] >> R1;
        //fs["R2"] >> R2;
        //fs["P1"] >> P1;
        //fs["P2"] >> P2;
        //fs["Q"]  >> Q;
		
        stereoRectify( M1, D1, M2, D2, img_size, R, T, R1, R2, P1, P2, Q, CALIB_ZERO_DISPARITY, 0.0, img_size, &roi1, &roi2 );

		initUndistortRectifyMap(M1, D1, R1, P1, img_size, CV_16SC2, map11, map12);
    	initUndistortRectifyMap(M2, D2, R2, P2, img_size, CV_16SC2, map21, map22);
        
    }



    

    
    numberOfDisparities = numberOfDisparities > 0 ? numberOfDisparities : ((img_size.width/8) + 15) & -16;
    
    bm.state->roi1 = roi1;  
    bm.state->roi2 = roi2;  
    
    bm.state->preFilterSize = 5;
	bm.state->preFilterType = CV_STEREO_BM_NORMALIZED_RESPONSE;
//	bm.state->trySmallerWindows = 0;

	bm.state->preFilterCap = 31;
    bm.state->SADWindowSize = SADWindowSize > 0 ? SADWindowSize : 9;
    bm.state->minDisparity = -2;
    bm.state->numberOfDisparities = numberOfDisparities;
    bm.state->textureThreshold = 6;
    bm.state->uniquenessRatio = 8; 
    bm.state->speckleWindowSize = 0;
    bm.state->speckleRange = 0;
    bm.state->disp12MaxDiff = 1;
    
    sgbm.preFilterCap = 51;
    sgbm.SADWindowSize = SADWindowSize > 0 ? SADWindowSize : 3;
    
    int cn = 1;  // 1 channel gray images 
    
    sgbm.P1 = 8*cn*sgbm.SADWindowSize*sgbm.SADWindowSize;
    sgbm.P2 = 32*cn*sgbm.SADWindowSize*sgbm.SADWindowSize;
    sgbm.minDisparity = 0;
    sgbm.numberOfDisparities = numberOfDisparities;
    sgbm.uniquenessRatio = 5;
    sgbm.speckleWindowSize = bm.state->speckleWindowSize;
    sgbm.speckleRange = bm.state->speckleRange;
    sgbm.disp12MaxDiff = 1;
    sgbm.fullDP = alg == STEREO_HH;
    
    var.levels = 3;     	// ignored with USE_AUTO_PARAMS
    var.pyrScale = 0.5;	// ignored with USE_AUTO_PARAMS
    var.nIt = 25;
    var.minDisp = -numberOfDisparities;	
    var.maxDisp = 0;
    var.poly_n = 3;
    var.poly_sigma = 0.0;
    var.fi = 15.0f;
    var.lambda = 0.03f;
    var.penalization = var.PENALIZATION_TICHONOV;	// ignored with USE_AUTO_PARAMS
    var.cycle = var.CYCLE_V;			// ignored with USE_AUTO_PARAMS
    var.flags = var.USE_SMART_ID | var.USE_AUTO_PARAMS | var.USE_INITIAL_DISPARITY | var.USE_MEDIAN_FILTERING ;
    
    Mat disp, disp8;
        
    bool flgExit = false;

    namedWindow("Params", 0);

	int tb1_SAD = sgbm.SADWindowSize;
	createTrackbar("SAD win.","Params",&tb1_SAD, 61, 0);
	int tb2_prefilterCap = sgbm.preFilterCap;
	createTrackbar("Preflt.Cap","Params",&tb2_prefilterCap, 256, 0);
	int tb3_minDisparity = sgbm.minDisparity;
	createTrackbar("min. disp.","Params",&tb3_minDisparity, 256, 0);
	int tb4_numberOfDisparities = sgbm.numberOfDisparities;
	createTrackbar("num. dispar.","Params",&tb4_numberOfDisparities, 256, 0);
	int tb5_uniquenessRatio = sgbm.uniquenessRatio;
	createTrackbar("uniq. ratio","Params",&tb5_uniquenessRatio, 256, 0);
	int tb6_speckleWindowSize = sgbm.speckleWindowSize;
	createTrackbar("Spk. winsize","Params",&tb6_speckleWindowSize, 256, 0);
	int tb7_speckleRange = sgbm.speckleRange;
	createTrackbar("Spk. range","Params",&tb7_speckleRange, 256, 0);


    namedWindow("rectified", 1);
   // namedWindow("right", 1);
    namedWindow("disparity", 1);
    if(point_cloud_filename)
    {
	    namedWindow("depth map", 1);
    }

	Mat canvas;
	int w = img1.cols;
    int h = img1.rows;
    canvas.create(h, w*2, CV_8UC1);


    while(!flgExit ){
    
   	 if( scale != 1.f )
   	 {
        	int method = scale < 1 ? INTER_AREA : INTER_CUBIC;
        	resize(img1, temp1, Size(), scale, scale, method);
        	img1 = temp1;
        	resize(img2, temp2, Size(), scale, scale, method);
        	img2 = temp2;
    	}
    

		
    
        remap(img1, img1r, map11, map12, INTER_LINEAR);
        remap(img2, img2r, map21, map22, INTER_LINEAR);
        
        img1 = img1r;
        img2 = img2r;




    	if( alg == STEREO_BM )
        	bm(img1, img2, disp);
    	else if( alg == STEREO_VAR ) {
        	var(img1, img2, disp);
		}
    	else if( alg == STEREO_SGBM || alg == STEREO_HH )
		{   
			sgbm.SADWindowSize = tb1_SAD;
			// cn = 1  for 1 channel gray images 
			sgbm.P1 = 8*cn*sgbm.SADWindowSize*sgbm.SADWindowSize;
			sgbm.P2 = 32*cn*sgbm.SADWindowSize*sgbm.SADWindowSize;

			sgbm.preFilterCap = tb2_prefilterCap;		    
			sgbm.minDisparity = tb3_minDisparity;

			sgbm.numberOfDisparities =  tb4_numberOfDisparities % 16 == 0 ? tb4_numberOfDisparities : 
					16 * ((int) tb4_numberOfDisparities / 16);
			if (sgbm.numberOfDisparities < 16) 
					sgbm.numberOfDisparities = 16;
			setTrackbarPos("num. dispar.","Params", sgbm.numberOfDisparities);

			sgbm.uniquenessRatio = tb5_uniquenessRatio;
			sgbm.speckleWindowSize = tb6_speckleWindowSize;
			sgbm.speckleRange = tb7_speckleRange;
			//sgbm.disp12MaxDiff = 1;

			sgbm(img1, img2, disp); }
		else if (alg == STEREO_LEGACY)
		{
			IplImage leftIm = img1;
			IplImage rightIm = img2;
			IplImage* dispIm = cvCloneImage(&leftIm);
			cvZero(dispIm);
			cvFindStereoCorrespondence(&leftIm, &rightIm, CV_DISPARITY_BIRCHFIELD, dispIm, numberOfDisparities,
					50, 15, 3, 6, 8);

			disp = dispIm;
		}
//    	t = getTickCount() - t;
//    	printf("Time elapsed: %fms\n", t*1000/getTickFrequency());


	// Display the rectified images from cameras
        Mat canvasPart = canvas(Rect(w*0, 0, w, h));
        resize(img1, canvasPart, canvasPart.size(), 0, 0, CV_INTER_AREA);
        
        canvasPart = canvas(Rect(w*1, 0, w, h));
        resize(img2, canvasPart, canvasPart.size(), 0, 0, CV_INTER_AREA);

		for(int j = 0; j < canvas.rows; j += 16 )
             line(canvas, Point(0, j), Point(canvas.cols, j), Scalar(0, 255, 0), 1, 8);
        imshow("rectified", canvas);
  //      imshow("right", img2);

    
    	if(point_cloud_filename)
    	{
		
//        	printf("storing the point cloud...");
//        	fflush(stdout);
        	Mat xyz;
			reprojectImageTo3D(disp, xyz, Q, true, CV_32FC3);
//        	saveXYZ(point_cloud_filename, xyz);
//	        printf("\n");
		
			IplImage depth = xyz;
			double Zmin= 9999.0;
			double Zmax= 0.0;

			float* fptr= (float*)(depth.imageData) + 2; // pointer to first Z value 
			float fv;  // for 32F format

			IplImage* depthShow = cvCreateImage(cvGetSize(&depth), 8, 1);
			unsigned char *cptr = (unsigned char *)(depthShow->imageData);

			// search min and max
			for(int y=0; y<depth.height; y++)
			{
				for(int x=0; x<depth.width; x++)
				{
					fv = *fptr;  // read the depth
					if ((fv < MAX_DEPTH_Z) && (fv >= Zmax)) // too far
					{	Zmax = fv; }
					else {if(fv < Zmin)    // too close
							Zmin = fv; }
					fptr+= 3;
				}
			}

			// coefficients for linear conversion
			double A = 255.0 / (Zmin - Zmax);
			double B = - 255.0 * (Zmax / (Zmin - Zmax));

			//LOOP
			fptr= (float*)(depth.imageData) + 2; 
			for(int y=0; y<depth.height; y++)
			{
				for(int x=0; x<depth.width; x++)
				{
					fv = *fptr;  // read the depth
					if (fv >= MAX_DEPTH_Z) // too far
					{	*cptr++ = 0; }
					else 
					{ if(fv < Zmin)    // too close
							{	*cptr++ = 255; }
						  else 
							{*cptr++ = (unsigned char)( (double)fv * A + B );}      
					}
					
					fptr+= 3;
				}
			}
			cvShowImage("depth map", depthShow);
			cvReleaseImage(&depthShow);
    	}
	// conversion for showing
	if( alg != STEREO_VAR )
    	disp.convertTo(disp8, CV_8U, 255/(numberOfDisparities*16.));
	else
    	disp.convertTo(disp8, CV_8U);
    imshow("disparity", disp8);
	
	// Handle pause/unpause and ESC
	int c = cvWaitKey( 15 );
	if( c == 'p' ){
		c = 0;
		while( c != 'p' && c != 27 ){
			c = cvWaitKey( 250 );
		}
	}
	if( c == 27 )
		flgExit = true;
		
	// Get next couple of images
	cvGrabFrame( captureL );
	cvGrabFrame( captureR );

	imageL = cvRetrieveFrame( captureL );
	imageR = cvRetrieveFrame( captureR );
	cvCvtColor(imageL, grayL, CV_BGR2GRAY);
	cvCvtColor(imageR, grayR, CV_BGR2GRAY);
	img1 = grayL;
   	img2 = grayR;
    } // End camera while loop
    
	
    cvReleaseCapture(&captureL);
    cvReleaseCapture(&captureR);
    
    return 0;
}
